<?php

namespace XF2SEO\SEOPackage\XF\BbCode\Renderer;

use XF\BbCode\RuleSet;
use XF2SEO\SEOPackage\App;
use XF2SEO\SEOPackage\Entity\Keyword;
use XF\Mvc\Entity\AbstractCollection;

class Html extends XFCP_Html
{
    /**
     * @var array
     */
    private $seoPackageTooltipElements = [];
    /**
     * @var array
     */
    private $seoPackageReplaces = [];
    /**
     * @var array
     */
    private $seoPackageKeywordRenderCounts = [];
    /**
     * @var null|AbstractCollection
     */
    private $seoPackageKeywords;
    /**
     * @var int
     */
    private static $seoPackageCounters = 0;

    /**
     * @param array $ast
     * @param RuleSet $rules
     * @param array $options
     * @return string
     */
    public function renderAst(array $ast, RuleSet $rules, array $options = [])
    {
        $options[App::RENDER_OPTION_KEYWORDS] = true;

        $output = parent::renderAst($ast, $rules, $options);
        $output .= implode('', $this->seoPackageTooltipElements);

        return $output;
    }

    /**
     * @param array $tree
     * @param array $options
     * @return string
     */
    public function renderSubTree(array $tree, array $options)
    {
        if (isset($options[App::RENDER_OPTION_KEYWORDS])) {
            foreach ($tree as &$element) {
                if (is_array($element)) {
                    // not render in tag
                    continue;
                }

                $element = App::TEXT_PLACEHOLDER . $element;
            }
            unset($element);

            unset($options[App::RENDER_OPTION_KEYWORDS]);
        }

        return parent::renderSubTree($tree, $options);
    }

    /**
     * @param mixed $string
     * @param array $options
     * @return mixed
     */
    public function renderString($string, array $options)
    {
        $plainText = false;
        if (strpos($string, App::TEXT_PLACEHOLDER) === 0) {
            $string = substr($string, strlen(App::TEXT_PLACEHOLDER));
            $plainText = true;
        }

        $rendered = parent::renderString($string, $options);

        self::$seoPackageCounters++;
        if ($plainText) {
            $segments = explode("\n", $rendered);
            $partRendered = [];
            foreach ($segments as $segment) {
                $partRendered[] = $this->renderSEOPackageKeywordInPlain($segment, $options);
            }
            $_rendered = implode("\n", $partRendered);
            $holder = substr(App::TEXT_PLACEHOLDER, 0, -1) . self::$seoPackageCounters . "\0";
            if ($rendered !== $_rendered) {
                $this->seoPackageReplaces[$holder] = $_rendered;

                $rendered = $holder;
            }
        }

        return $rendered;
    }

    /**
     * @param mixed $output
     * @return string
     */
    public function filterFinalOutput($output)
    {
        foreach ($this->seoPackageReplaces as $find => $replace) {
            $output = str_replace($find, $replace, $output);
        }

        return parent::filterFinalOutput($output);
    }

    /**
     * @param mixed $string
     * @param array $options
     * @return mixed
     */
    public function filterString($string, array $options)
    {
        $prefix = substr(App::TEXT_PLACEHOLDER, 0, -1);
        if (strpos($string, $prefix) === 0) {
            return $string;
        }

        return parent::filterString($string, $options);
    }

    protected function getSEOPackageKeywordList(): AbstractCollection
    {
        if ($this->seoPackageKeywords === null) {
            $this->seoPackageKeywords = \XF::finder('XF2SEO\SEOPackage:Keyword')
                ->where('active', 1)
                ->order('execute_order')
                ->fetch();
        }

        return $this->seoPackageKeywords;
    }

    protected function renderSEOPackageKeywordInPlain(string $string, array $options = []): string
    {
        if (strlen($string) === 0) {
            return $string;
        }

        $keywords = $this->getSEOPackageKeywordList();
        if ($keywords->count() === 0) {
            return $string;
        }
        $replaceMap = [];

        /** @var Keyword $keyword */
        foreach ($keywords as $keyword) {
            if (strlen($keyword->replacement) === 0 && strlen($keyword->target_url) === 0) {
                continue;
            }

            $replaceLimit = $keyword->options['replace_limit'] ?? 0;
            $replaceTotal = $this->seoPackageKeywordRenderCounts[$keyword->keyword_id] ?? 0;
            preg_match_all(
                '#\b' . preg_quote($keyword->keyword, '#') . '\b#ui',
                $string,
                $matches,
                PREG_OFFSET_CAPTURE | PREG_SET_ORDER
            );
            if (count($matches) === 0) {
                continue;
            }

            $tree = [];
            $lastOffset = 0;
            foreach ($matches[0] as $match) {
                $tree[] = substr($string, $lastOffset, $match[1]);
                $tree[] = [$match[0]];

                $lastOffset += $match[1] + strlen($match[0]);
            }
            $tree[] = substr($string, $lastOffset);
            $replacement = $this->getSEOPackageReplacement($keyword);
            $output = '';
            foreach ($tree as $element) {
                if (is_string($element)) {
                    $output .= $element;

                    continue;
                }

                if ($replaceLimit > 0 && $replaceTotal >= $replaceLimit) {
                    $output .= $element[0];

                    continue;
                }

                $linkText = $replacement[0];
                $link = $replacement[1];
                if ($linkText === null && $link === null) {
                    $output .= $element[0];

                    continue;
                }

                self::$seoPackageCounters++;
                $holder = "\0{$keyword['keyword_id']}_" . self::$seoPackageCounters . "\0";

                if ($link !== null) {
                    $link = str_replace(App::TEXT_PLACEHOLDER, $element[0], $link);
                    $replaceMap[$holder] = $link;
                } else {
                    $replaceMap[$holder] = $linkText;
                }
                $output .= $holder;
                $replaceTotal++;
            }

            $string = $output;
            $this->seoPackageKeywordRenderCounts[$keyword->keyword_id] = $replaceTotal;
        }

        foreach ($replaceMap as $holder => $replace) {
            $string = str_replace($holder, $replace, $string);
        }

        return $string;
    }

    protected function getSEOPackageReplacement(Keyword $keyword): array
    {
        if (strlen($keyword->replacement) > 0) {
            $linkText = $keyword->replacement;
        } else {
            $linkText = null;
        }

        $targetUrl = $keyword->target_url;
        $link = null;

        if (strlen($targetUrl) > 0) {
            $linkInfo = $this->formatter->getLinkClassTarget($targetUrl);

            $link = $this->getRenderedLink($linkText ?? App::TEXT_PLACEHOLDER, $targetUrl, $linkInfo);
            $link = trim($link);

            if (strlen($keyword->description) > 0) {
                $link = substr($link, 0, 3)
                    . 'data-xf-init="element-tooltip" data-html="true" data-shortcut="keyword'
                    . $keyword->keyword_id . '-description" '
                    . 'data-element="< .message-body | .keyword' . $keyword->keyword_id . '-description" '
                    . substr($link, 3);
                $this->seoPackageTooltipElements[$keyword->keyword_id] = '<div class="keyword'
                    . $keyword->keyword_id . '-description node-description--tooltip js-nodeDescTooltip" style="display:none">'
                    . $keyword->description
                . '</div>';
            }
        }

        return [$linkText, $link];
    }
}
